////////////////////////////////////////////////////////////
// BmpZoomer
// Written by Manuke 2001

// Include

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <memory.h>
#include <math.h>

#include <windows.h>

// JPEG

#ifdef FILE_SUPPORT_JPEG

#define XMD_H
#undef FAR
typedef short INT16;

extern "C" {
#include "jinclude.h"
#include "jpeglib.h"
#include "jerror.h"
};

#endif	// FILE_SUPPORT_JPEG

// PNG

#ifdef FILE_SUPPORT_PNG

#include "png.h"

#endif	// FILE_SUPPORT_PNG

// define

#ifndef _CONSOLE

#define main xmain

#endif

// Constant

const double PI = 3.14159265358979323846;

// Color Class

class CColor {
// Attribute
public:
	// Value
	double fRed, fGreen, fBlue;

// Construct
public:
	// Default Constructor
	CColor() :
		fRed(0),
		fGreen(0),
		fBlue(0)
	{
	}
	// Copy Constructor
	CColor(const CColor& colOrg) :
		fRed(colOrg.fRed),
		fGreen(colOrg.fGreen),
		fBlue(colOrg.fBlue)
	{
	}
	// COLORREF Constructor
	CColor(COLORREF colOrg) :
		fRed(GetRValue(colOrg)/255.0),
		fGreen(GetGValue(colOrg)/255.0),
		fBlue(GetBValue(colOrg)/255.0)
	{
	}

// Operator
public:
	// Let
	CColor& operator=(const CColor& colOther) {
		fRed = colOther.fRed;
		fGreen = colOther.fGreen;
		fBlue = colOther.fBlue;
		return *this;
	}
	// Add Let
	CColor& operator+=(const CColor& colOther) {
		fRed += colOther.fRed;
		fGreen += colOther.fGreen;
		fBlue += colOther.fBlue;
		return *this;
	}
	// Sub Let
	CColor& operator-=(const CColor& colOther) {
		fRed -= colOther.fRed;
		fGreen -= colOther.fGreen;
		fBlue -= colOther.fBlue;
		return *this;
	}
	// Mul Let
	CColor& operator*=(double fValue) {
		fRed *= fValue;
		fGreen *= fValue;
		fBlue *= fValue;
		return *this;
	}
	// Div Let
	CColor& operator/=(double fValue) {
		fRed /= fValue;
		fGreen /= fValue;
		fBlue /= fValue;
		return *this;
	}
	// Add
	CColor operator+(const CColor& colOther) const {
		CColor colNew = *this;
		colNew += colOther;
		return colNew;
	}
	// Sub
	CColor operator-(const CColor& colOther) const {
		CColor colNew = *this;
		colNew -= colOther;
		return colNew;
	}
	// Mul
	CColor operator*(double fValue) const {
		CColor colNew = *this;
		colNew *= fValue;
		return colNew;
	}
	// Div
	CColor operator/(double fValue) const {
		CColor colNew = *this;
		colNew /= fValue;
		return colNew;
	}
	// COLORREF cast
	operator COLORREF() const {
		int nRed = (int)(fRed*255+0.5),
			nGreen = (int)(fGreen*255+0.5),
			nBlue = (int)(fBlue*255+0.5);
		if (nRed < 0) {
			nRed = 0;
		} else if (nRed > 255) {
			nRed = 255;
		}
		if (nGreen < 0) {
			nGreen = 0;
		} else if (nGreen > 255) {
			nGreen = 255;
		}
		if (nBlue < 0) {
			nBlue = 0;
		} else if (nBlue > 255) {
			nBlue = 255;
		}
		return RGB(nRed, nGreen, nBlue);
	}
};

// Show Message

void xprintf(const char* pszFormat, ...) {

#ifdef _CONSOLE

	va_list ap;
	va_start(ap, pszFormat);
	vfprintf(stderr, pszFormat, ap);
	va_end(ap);

#else

	pszFormat;

#endif

}

// Temporary Window

LRESULT CALLBACK TmpWndProc(
	HWND hwnd, UINT nMessage, WPARAM wParam, LPARAM lParam)
{
	return DefWindowProc(hwnd, nMessage, wParam, lParam);
}

HWND CreateTmpWindow() {
	HINSTANCE hInstance = (HINSTANCE)GetModuleHandle(NULL);
	WNDCLASS wc;
	memset(&wc, 0, sizeof(wc));
	wc.style = CS_VREDRAW | CS_HREDRAW;
	wc.lpfnWndProc = TmpWndProc;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName = NULL;
	wc.lpszClassName = "Tmp";
	if (RegisterClass(&wc) == NULL) {
		return NULL;
	}
	return CreateWindow(
		"Tmp", "Tmp",
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, hInstance, NULL);
}

// Bitmap Function

bool GetBitmapSize(HBITMAP hbmpTarget, int& nWidth, int& nHeight) {
	BITMAP bm;
	memset(&bm, 0, sizeof(bm));
	if (GetObject(hbmpTarget, sizeof(bm), &bm) != sizeof(bm)) {
		return false;
	}
	if ((bm.bmWidth <= 0) || (bm.bmHeight == 0)) {
		return false;
	}
	nWidth = bm.bmWidth;
	nHeight = abs(bm.bmHeight);
	return true;
}

bool IsBitmapGrayScale(HBITMAP hbmpTarget) {
	DIBSECTION ds;
	memset(&ds, 0, sizeof(ds));
	if (GetObject(hbmpTarget, sizeof(ds), &ds) != sizeof(ds)) {
		return false;
	}
	int nDepth = ds.dsBmih.biBitCount*ds.dsBmih.biPlanes,
		nPalCount = max(
			ds.dsBmih.biClrUsed,
			ds.dsBmih.biClrImportant);
	if ((nPalCount == 0) && (nDepth <= 8)) {
		nPalCount = (1 << nDepth);
	}
	if ((nDepth != 8) || (nPalCount != 256)) {
		return false;
	}
	bool bGrayScale = false;
	RGBQUAD argbPalettes[256];
	memset(argbPalettes, 0, sizeof(argbPalettes));
	HDC hdcMem = CreateCompatibleDC(NULL);
	if (hdcMem != NULL) {
		HBITMAP hbmpOrg = (HBITMAP)SelectObject(hdcMem, hbmpTarget);
		int nPalCount = GetDIBColorTable(
			hdcMem, 0, 256, argbPalettes);
		SelectObject(hdcMem, hbmpOrg);
		DeleteDC(hdcMem);
		if (nPalCount == 256) {
			bGrayScale = true;
			for (int i = 0; i < 256; i++) {
				if ((argbPalettes[i].rgbRed != i) ||
					(argbPalettes[i].rgbGreen != i) ||
					(argbPalettes[i].rgbBlue != i))
				{
					bGrayScale = false;
					break;
				}
			}
		}
	}
	return bGrayScale;
}

bool LoadBitmapFileFromBMP(
	const char* pszFileName, HBITMAP& hbmpLoad,
	bool& bGrayScale)
{
	bool bResult;
	bGrayScale = false;
	hbmpLoad = (HBITMAP)LoadImage(
		NULL, pszFileName,
		IMAGE_BITMAP, 0, 0,
		LR_LOADFROMFILE | LR_CREATEDIBSECTION);
	if (hbmpLoad != NULL) {
		bResult = true;
	} else {
		FILE* pFile = NULL;
		HDC hdcDesktop = NULL;
		BYTE* pbtBits = NULL;
		try {
			pFile = fopen(pszFileName, "rb");
			if (pFile == NULL) {
				throw 1;
			}
			BITMAPFILEHEADER bmfh;
			if (fread(&bmfh, 1, sizeof(bmfh), pFile) !=
					sizeof(bmfh))
			{
				throw 2;
			}
			if (bmfh.bfType != (WORD)('B' | ('M' << 8))) {
				throw 3;
			}
			BYTE abtDIB[sizeof(BITMAPINFOHEADER)+
				sizeof(DWORD)*3+sizeof(RGBQUAD)*256];
			BYTE* pbtDIB = abtDIB;
			if (fread(pbtDIB, 1, sizeof(BITMAPINFOHEADER), pFile) !=
					sizeof(BITMAPINFOHEADER))
			{
				throw 3;
			}
			BITMAPINFO* pbmi = (BITMAPINFO*)pbtDIB;
			pbtDIB += sizeof(BITMAPINFOHEADER);
			if (pbmi->bmiHeader.biSize < sizeof(BITMAPINFOHEADER)) {
				throw 4;
			}
			if (pbmi->bmiHeader.biSize > sizeof(BITMAPINFOHEADER)) {
				if (fseek(
						pFile,
						pbmi->bmiHeader.biSize-sizeof(BITMAPINFOHEADER),
						SEEK_CUR) != 0)
				{
					throw 13;
				}
				pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
			}
			if ((pbmi->bmiHeader.biCompression != BI_RGB) &&
				(pbmi->bmiHeader.biCompression != BI_BITFIELDS))
			{
				throw 5;
			}
			int nWidth = pbmi->bmiHeader.biWidth,
				nHeight = pbmi->bmiHeader.biHeight,
				nDepth = pbmi->bmiHeader.biBitCount*
					pbmi->bmiHeader.biPlanes,
				nScanLine = ((nWidth*nDepth+31)/32)*4,
				nBitsLength = nScanLine*nHeight,
				nPalCount = max(
					pbmi->bmiHeader.biClrUsed,
					pbmi->bmiHeader.biClrImportant);
			if ((nPalCount == 0) && (nDepth <= 8)) {
				nPalCount = (1 << nDepth);
			}
			if (pbmi->bmiHeader.biCompression == BI_BITFIELDS) {
				if (fread(pbtDIB, 1, sizeof(DWORD)*3, pFile) !=
						sizeof(DWORD)*3)
				{
					throw 6;
				}
				pbtDIB += sizeof(DWORD)*3;
			}
			if (nPalCount > 0) {
				if (fread(pbtDIB, 1, sizeof(RGBQUAD)*nPalCount, pFile) !=
						sizeof(RGBQUAD)*nPalCount)
				{
					throw 8;
				}
			}
			pbtBits = new BYTE[nBitsLength];
			if (pbtBits == NULL) {
				throw 9;
			}
			if (fread(pbtBits, 1, nBitsLength, pFile) !=
					(size_t)nBitsLength)
			{
				throw 10;
			}
			hdcDesktop = GetDC(HWND_DESKTOP);
			if (hdcDesktop == NULL) {
				throw 11;
			}
			void* pvBits;
			hbmpLoad = CreateDIBSection(
				hdcDesktop,
				pbmi, DIB_RGB_COLORS,
				&pvBits,
				NULL, 0);
			if (hbmpLoad == NULL) {
				throw 12;
			}
			memcpy(pvBits, pbtBits, nBitsLength);
		} catch (...) {
		}
		if (pFile != NULL) {
			fclose(pFile);
		}
		if (hdcDesktop != NULL) {
			ReleaseDC(HWND_DESKTOP, hdcDesktop);
		}
		if (pbtBits != NULL) {
			delete [] pbtBits;
		}
		bResult = (hbmpLoad != NULL);
	}
	if (bResult) {
		bGrayScale = IsBitmapGrayScale(hbmpLoad);
	}
	return bResult;
}

bool SaveBitmapFileToBMP(
	const char* pszFileName, HBITMAP hbmpSave,
	bool /*bGrayScale*/, int /*nQuality*/)
{
	DIBSECTION ds;
	memset(&ds, 0, sizeof(ds));
	if (GetObject(hbmpSave, sizeof(ds), &ds) != sizeof(ds)) {
		return false;
	}
	int nWidth = ds.dsBmih.biWidth,
		nHeight = ds.dsBmih.biHeight,
		nDepth = ds.dsBmih.biBitCount*ds.dsBmih.biPlanes,
		nScanLine = ((nWidth*nDepth+31)/32)*4,
		nBitsLength = nScanLine*nHeight,
		nPalCount = max(
			ds.dsBmih.biClrUsed,
			ds.dsBmih.biClrImportant);
	if ((nPalCount == 0) && (nDepth <= 8)) {
		nPalCount = (1 << nDepth);
	}
	RGBQUAD argbPalettes[256];
	memset(argbPalettes, 0, sizeof(argbPalettes));
	if (nPalCount > 0) {
		HDC hdcMem = CreateCompatibleDC(NULL);
		if (hdcMem != NULL) {
			HBITMAP hbmpOrg = (HBITMAP)SelectObject(hdcMem, hbmpSave);
			GetDIBColorTable(hdcMem, 0, nPalCount, argbPalettes);
			SelectObject(hdcMem, hbmpOrg);
			DeleteDC(hdcMem);
		}
	}
	int nInfoLength = sizeof(BITMAPINFOHEADER)+
		((ds.dsBmih.biCompression == BI_BITFIELDS)?
			(sizeof(DWORD)*3): 0)+
		sizeof(RGBQUAD)*nPalCount;
	BITMAPFILEHEADER bmfh;
	memset(&bmfh, 0, sizeof(bmfh));
	bmfh.bfType = (WORD)('B' | ('M' << 8));
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER)+nInfoLength;
	bmfh.bfSize = bmfh.bfOffBits+nBitsLength;
	bool bResult = false;
	FILE* pFile = fopen(pszFileName, "wb");
	if (pFile != NULL) {
		if ((fwrite(&bmfh, 1, sizeof(bmfh), pFile) !=
				sizeof(bmfh)) ||
			(fwrite(&ds.dsBmih, 1, sizeof(ds.dsBmih), pFile) !=
				sizeof(ds.dsBmih)))
		{
			// Error
		} else if (
			(ds.dsBmih.biCompression == BI_BITFIELDS) &&
			(fwrite(ds.dsBitfields, 1, sizeof(DWORD)*3, pFile) !=
				sizeof(DWORD)*3))
		{
			// Error
		} else if (
			(nPalCount > 0) &&
			(fwrite(argbPalettes, 1, sizeof(RGBQUAD)*nPalCount, pFile) !=
				sizeof(RGBQUAD)*nPalCount))
		{
			// Error
		} else if (
			(fwrite(ds.dsBm.bmBits, 1, nBitsLength, pFile) !=
				(size_t)nBitsLength) ||
			(fflush(pFile) != 0))
		{
			// Error
		} else {
			bResult = true;
		}
		fclose(pFile);
		if (!bResult) {
			_unlink(pszFileName);
		}
	}
	return bResult;
}

#ifdef FILE_SUPPORT_JPEG

void JPEG_error_exit(j_common_ptr cinfo) {
	char szBuffer[JMSG_LENGTH_MAX];
	(*cinfo->err->format_message)(cinfo, szBuffer);
	xprintf("%s\n", szBuffer);
	jpeg_destroy(cinfo);
	throw "internal error";
}

bool IsFileNameJPEG(const char* pszFileName) {
	int nLength = strlen(pszFileName);
	return
		((nLength >= 4) &&
			(stricmp(pszFileName+nLength-4, ".jpg") == 0)) ||
		((nLength >= 5) &&
			(stricmp(pszFileName+nLength-5, ".jpeg") == 0));
}

bool LoadBitmapFileFromJPEG(
	const char* pszFileName, HBITMAP& hbmpLoad,
	bool& bGrayScale)
{
	bool bResult = true;
	FILE* pFile = NULL;
	hbmpLoad = NULL;
	bGrayScale = false;
	jpeg_decompress_struct cinfo;
	bool cinfo_created = false;
	jpeg_error_mgr jerr;
	cinfo.err = jpeg_std_error(&jerr);
	jerr.error_exit = JPEG_error_exit;
	try {
		pFile = fopen(pszFileName, "rb");
		if (pFile == NULL) {
			throw "cannot open file";
		}
		jpeg_create_decompress(&cinfo);
		cinfo_created = true;
		jpeg_stdio_src(&cinfo, pFile);
		jpeg_read_header(&cinfo, TRUE);
		jpeg_start_decompress(&cinfo);
		int nBitCount = 24;
		if (cinfo.out_color_space == JCS_GRAYSCALE) {
			bGrayScale = true;
			nBitCount = 8;
		} else if (cinfo.out_color_space == JCS_RGB) {
			nBitCount = cinfo.quantize_colors? 8: 24;
		} else {
			throw "unknown format";
		}
		BYTE abtTmp[sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*256];
		memset(abtTmp, 0, sizeof(abtTmp));
		BITMAPINFO* pbmi = (BITMAPINFO*)abtTmp;
		pbmi->bmiHeader.biSize     = sizeof(BITMAPINFOHEADER);
		pbmi->bmiHeader.biWidth    = cinfo.output_width;
		pbmi->bmiHeader.biHeight   = cinfo.output_height;
		pbmi->bmiHeader.biBitCount = (WORD)nBitCount;
		pbmi->bmiHeader.biPlanes   = 1;
		JSAMPARRAY colormap = cinfo.colormap;
		if (nBitCount > 8) {
			// No Palette
		} else if (colormap != NULL) {
			int num_colors = cinfo.actual_number_of_colors;
			if (num_colors > 256) {
				throw "color palette over 256";
			}
			if (cinfo.out_color_components == 3) {
				for (int i = 0; i < num_colors; i++) {
					pbmi->bmiColors[i].rgbRed =
						(BYTE)GETJSAMPLE(colormap[0][i]);
					pbmi->bmiColors[i].rgbGreen =
						(BYTE)GETJSAMPLE(colormap[1][i]);
					pbmi->bmiColors[i].rgbBlue =
						(BYTE)GETJSAMPLE(colormap[2][i]);
				}
			} else {
				for (int i = 0; i < num_colors; i++) {
					pbmi->bmiColors[i].rgbRed =
						(BYTE)GETJSAMPLE(colormap[0][i]);
					pbmi->bmiColors[i].rgbGreen =
						(BYTE)GETJSAMPLE(colormap[0][i]);
					pbmi->bmiColors[i].rgbBlue =
						(BYTE)GETJSAMPLE(colormap[0][i]);
				}
			}
		} else {
			for (int i = 0; i < 256; i++) {
				pbmi->bmiColors[i].rgbRed   = (BYTE)i;
				pbmi->bmiColors[i].rgbGreen = (BYTE)i;
				pbmi->bmiColors[i].rgbBlue  = (BYTE)i;
			}
		}
		void* pvBits = NULL;
		HDC hdcDesktop = GetDC(GetDesktopWindow());
		if (hdcDesktop != NULL) {
			hbmpLoad = (HBITMAP)CreateDIBSection(
				hdcDesktop, pbmi, DIB_RGB_COLORS, &pvBits, NULL, 0);
			ReleaseDC(GetDesktopWindow(), hdcDesktop);
		}
		if (hbmpLoad == NULL) {
			throw "cannot create dib";
		}
		int row_stride = cinfo.output_width*cinfo.output_components;
		JSAMPARRAY buffer = (*cinfo.mem->alloc_sarray)(
			(j_common_ptr)&cinfo, JPOOL_IMAGE, row_stride, 1);
		int nScanLine = ((cinfo.output_width*nBitCount+31)/32)*4;
		BYTE* pbtScanLine = (BYTE*)pvBits+
			(cinfo.output_height-1)*nScanLine;
		while (cinfo.output_scanline < cinfo.output_height) {
			jpeg_read_scanlines(&cinfo, buffer, 1);
			if (nBitCount <= 8) {
				memcpy(pbtScanLine, buffer[0], nScanLine);
			} else {
				BYTE* pbtScanLine2 = pbtScanLine;
				JSAMPROW ptr = buffer[0];
				for (int i = cinfo.output_width; i > 0; i--) {
					pbtScanLine2[2] = (BYTE)GETJSAMPLE(*ptr++);
					pbtScanLine2[1] = (BYTE)GETJSAMPLE(*ptr++);
					pbtScanLine2[0] = (BYTE)GETJSAMPLE(*ptr++);
					pbtScanLine2 += 3;
				}
			}
			pbtScanLine -= nScanLine;
		}
		jpeg_finish_decompress(&cinfo);
		jpeg_destroy_decompress(&cinfo);
	} catch (const char* pszError) {
		xprintf("jpeg load error: %s\n", pszError);
		if (cinfo_created) {
			jpeg_destroy_decompress(&cinfo);
		}
		if (hbmpLoad != NULL) {
			DeleteObject(hbmpLoad);
			hbmpLoad = NULL;
		}
		bResult = false;
	}
	if (pFile != NULL) {
		fclose(pFile);
	}
	return bResult;
}

bool SaveBitmapFileToJPEG(
	const char* pszFileName, HBITMAP hbmpSave,
	bool bGrayScale, int nQuality)
{
	bool bResult = true;
	FILE* pFile = NULL;
	HDC hdcMem = NULL;
	HBITMAP hbmpOrg = NULL;
	jpeg_compress_struct cinfo;
	bool cinfo_created = false;
	jpeg_error_mgr jerr;
	cinfo.err = jpeg_std_error(&jerr);
	jerr.error_exit = JPEG_error_exit;
	try {
		pFile = fopen(pszFileName, "wb");
		if (pFile == NULL) {
			throw "cannot create file";
		}
		DIBSECTION ds;
		if (GetObject(hbmpSave, sizeof(ds), &ds) != sizeof(ds)) {
			throw "cannot get dibsection";
		}
		hdcMem = CreateCompatibleDC(NULL);
		if (hdcMem == NULL) {
			throw "cannot create device-context";
		}
		hbmpOrg = (HBITMAP)SelectObject(hdcMem, hbmpSave);
		jpeg_create_compress(&cinfo);
		cinfo_created = true;
		cinfo.image_width = ds.dsBmih.biWidth;
		cinfo.image_height = ds.dsBmih.biHeight;
		cinfo.input_components = 3;
		cinfo.in_color_space = JCS_RGB;
		jpeg_set_defaults(&cinfo);
		if (bGrayScale) {
			jpeg_set_colorspace(&cinfo, JCS_GRAYSCALE);
		}
		jpeg_set_quality(&cinfo, nQuality, TRUE);
		jpeg_stdio_dest(&cinfo, pFile);
		jpeg_start_compress(&cinfo, TRUE);
		int row_stride = cinfo.image_width*3;
		JSAMPARRAY buffer = (*cinfo.mem->alloc_sarray)(
			(j_common_ptr)&cinfo, JPOOL_IMAGE, row_stride, 1);
		while (cinfo.next_scanline < cinfo.image_height) {
			JSAMPROW ptr = buffer[0];
			for (unsigned x = 0; x < cinfo.image_width; x++) {
				COLORREF col = GetPixel(hdcMem, x, cinfo.next_scanline);
				*ptr++ = GetRValue(col);
				*ptr++ = GetGValue(col);
				*ptr++ = GetBValue(col);
			}
			jpeg_write_scanlines(&cinfo, buffer, 1);
		}
		jpeg_finish_compress(&cinfo);
		jpeg_destroy_compress(&cinfo);
	} catch (const char* pszError) {
		xprintf("jpeg save error: %s\n", pszError);
		if (cinfo_created) {
			jpeg_destroy_compress(&cinfo);
		}
		bResult = false;
	}
	if (pFile != NULL) {
		fclose(pFile);
		if (!bResult) {
			_unlink(pszFileName);
		}
	}
	if (hdcMem != NULL) {
		if (hbmpOrg != NULL) {
			SelectObject(hdcMem, hbmpOrg);
		}
		DeleteDC(hdcMem);
	}
	return bResult;
}

#endif	// FILE_SUPPORT_JPEG

#ifdef FILE_SUPPORT_PNG

void PNG_user_error_fn(
	png_structp /*png_ptr*/,
	png_const_charp error_msg)
{
	xprintf("%s\n", error_msg);
	throw "internal error";
}

// x

void PNG_user_warning_fn(
	png_structp /*png_ptr*/,
	png_const_charp /*warning_msg*/)
{
}

bool IsFileNamePNG(const char* pszFileName) {
	int nLength = strlen(pszFileName);
	return
		((nLength >= 4) &&
			(stricmp(pszFileName+nLength-4, ".png") == 0));
}

bool LoadBitmapFileFromPNG(
	const char* pszFileName, HBITMAP& hbmpLoad,
	bool& bGrayScale)
{
	bool bResult = true;
	FILE* pFile = NULL;
	hbmpLoad = NULL;
	bGrayScale = false;
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	RGBQUAD argbColors[256];
	png_uint_32 width, height;
	int bit_depth, color_type, interlace_type;
	png_bytep row_buf = NULL;
	try {
		pFile = fopen(pszFileName, "rb");
		if (pFile == NULL) {
			throw "cannot open file";
		}
		png_ptr = png_create_read_struct(
			PNG_LIBPNG_VER_STRING,
			NULL,
			PNG_user_error_fn,
			PNG_user_warning_fn);
		if (png_ptr == NULL) {
			throw "cannot create read struct";
		}
		info_ptr = png_create_info_struct(png_ptr);
		if (info_ptr == NULL) {
			throw "cannot create info struct";
		}
		png_init_io(png_ptr, pFile);
		png_read_info(png_ptr, info_ptr);
		png_get_IHDR(
			png_ptr, info_ptr,
			&width, &height,
			&bit_depth, &color_type, &interlace_type,
			NULL, NULL);
		bGrayScale =
			(color_type == PNG_COLOR_TYPE_GRAY) ||
			(color_type == PNG_COLOR_TYPE_GRAY_ALPHA);
		if (bit_depth > 8) {
			png_set_strip_16(png_ptr);
		}
		if (bit_depth < 8) {
			png_set_packing(png_ptr);
		}
		if ((color_type & PNG_COLOR_MASK_ALPHA) != 0) {
			png_set_strip_alpha(png_ptr);
		}
		if (bGrayScale && (bit_depth < 8)) {
			png_set_expand(png_ptr);
		}
		memset(argbColors, 0, sizeof(argbColors));
		if (bGrayScale) {
			for (int i = 0; i < 256; i++) {
				argbColors[i].rgbRed =
					argbColors[i].rgbGreen =
					argbColors[i].rgbBlue = (BYTE)i;
			}
		} else {
			png_colorp palette;
			int num_palette;
			if (png_get_PLTE(
					png_ptr, info_ptr, &palette, &num_palette))
			{
				if (num_palette > 256) {
					num_palette = 256;
				}
				for (int i = 0; i < num_palette; i++) {
					argbColors[i].rgbRed = palette[i].red;
					argbColors[i].rgbGreen = palette[i].green;
					argbColors[i].rgbBlue = palette[i].blue;
				}
			}
		}
		png_set_bgr(png_ptr);
		int number_passes = png_set_interlace_handling(png_ptr);
		void* pvBits = NULL;
		int nBitCount, nScanLine, nWidthBits;
		{ // dummy block
			nBitCount =
				(!bGrayScale && ((color_type & PNG_COLOR_MASK_PALETTE) == 0))?
					24: 8;
			nScanLine = ((width*nBitCount+31)/32)*4;
			nWidthBits = (width*nBitCount+7)/8;
			BYTE abtTmp[sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*256];
			memset(abtTmp, 0, sizeof(abtTmp));
			BITMAPINFO* pbmi = (BITMAPINFO*)abtTmp;
			pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
			pbmi->bmiHeader.biWidth = width;
			pbmi->bmiHeader.biHeight = height;
			pbmi->bmiHeader.biBitCount =(WORD)nBitCount;
			pbmi->bmiHeader.biPlanes = 1;
			memcpy(pbmi->bmiColors, argbColors, sizeof(argbColors));
			HDC hdcDesktop = GetDC(GetDesktopWindow());
			if (hdcDesktop != NULL) {
				hbmpLoad = (HBITMAP)CreateDIBSection(
					hdcDesktop, pbmi, DIB_RGB_COLORS, &pvBits, NULL, 0);
				ReleaseDC(GetDesktopWindow(), hdcDesktop);
			}
			if (hbmpLoad == NULL) {
				throw "cannot create dib";
			}
		}
		row_buf = (png_bytep)(new BYTE[nWidthBits]);
		for (int pass = 0; pass < number_passes; pass++) {
			BYTE* pbtBits = (BYTE*)pvBits+(height-1)*nScanLine;
			for (unsigned y = 0; y < height; y++) {
				memcpy(row_buf, pbtBits, nWidthBits);
				png_read_rows(png_ptr, &row_buf, NULL, 1);
				memcpy(pbtBits, row_buf, nWidthBits);
				pbtBits -= nScanLine;
			}
		}
		png_destroy_read_struct(
			&png_ptr,
			&info_ptr,
			(png_infopp)NULL);
	} catch (const char* pszError) {
		xprintf("png load error: %s\n", pszError);
		if (png_ptr != NULL) {
			png_destroy_read_struct(
				&png_ptr,
				(info_ptr != NULL)? &info_ptr: NULL,
				NULL);
		}
		if (hbmpLoad != NULL) {
			DeleteObject(hbmpLoad);
			hbmpLoad = NULL;
		}
		bResult = false;
	}
	if (pFile != NULL) {
		fclose(pFile);
	}
	if (row_buf != NULL) {
		delete [] (BYTE*)row_buf;
	}
	return bResult;
}

bool SaveBitmapFileToPNG(
	const char* pszFileName, HBITMAP hbmpSave,
	bool bGrayScale, int /*nQuality*/)
{
	bool bResult = true;
	FILE* pFile = NULL;
	HDC hdcMem = NULL;
	HBITMAP hbmpOrg = NULL;
	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;
	png_uint_32 width, height;
	png_bytep row_buf = NULL;
	try {
		pFile = fopen(pszFileName, "wb");
		if (pFile == NULL) {
			throw "cannot create file";
		}
		DIBSECTION ds;
		if (GetObject(hbmpSave, sizeof(ds), &ds) != sizeof(ds)) {
			throw "cannot get dibsection";
		}
		hdcMem = CreateCompatibleDC(NULL);
		if (hdcMem == NULL) {
			throw "cannot create device-context";
		}
		hbmpOrg = (HBITMAP)SelectObject(hdcMem, hbmpSave);
		png_ptr = png_create_write_struct(
			PNG_LIBPNG_VER_STRING,
			NULL,
			PNG_user_error_fn,
			PNG_user_warning_fn);
		if (png_ptr == NULL) {
			throw "cannot create write struct";
		}
		info_ptr = png_create_info_struct(png_ptr);
		if (info_ptr == NULL) {
			throw "cannot create info struct";
		}
		png_init_io(png_ptr, pFile);
		width = ds.dsBmih.biWidth;
		height = ds.dsBmih.biHeight;
		png_set_IHDR(
			png_ptr, info_ptr,
			width, height, 8,
			bGrayScale?
				PNG_COLOR_TYPE_GRAY:
				PNG_COLOR_TYPE_RGB,
			PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_BASE,
			PNG_FILTER_TYPE_BASE);
		png_write_info(png_ptr, info_ptr);
		png_set_bgr(png_ptr);
		row_buf = (png_bytep)(new BYTE[!bGrayScale? width*3: width]);
		for (unsigned y = 0; y < height; y++) {
			int i = 0;
			for (unsigned x = 0; x < width; x++) {
				COLORREF col = GetPixel(hdcMem, x, y);
				row_buf[i++] = GetBValue(col);
				if (!bGrayScale) {
					row_buf[i++] = GetGValue(col);
					row_buf[i++] = GetRValue(col);
				}
			}
			png_write_rows(png_ptr, &row_buf, 1);
		}
		png_write_end(png_ptr, info_ptr);
		png_destroy_write_struct(
			&png_ptr,
			&info_ptr);
	} catch (const char* pszError) {
		xprintf("png save error: %s\n", pszError);
		if (png_ptr != NULL) {
			png_destroy_write_struct(
				&png_ptr,
				(info_ptr != NULL)? &info_ptr: NULL);
		}
		bResult = false;
	}
	if (pFile != NULL) {
		fclose(pFile);
		if (!bResult) {
			_unlink(pszFileName);
		}
	}
	if (hdcMem != NULL) {
		if (hbmpOrg != NULL) {
			SelectObject(hdcMem, hbmpOrg);
		}
		DeleteDC(hdcMem);
	}
	if (row_buf != NULL) {
		delete [] (BYTE*)row_buf;
	}
	return bResult;
}

#endif	// FILE_SUPPORT_PNG

bool LoadBitmapFile(
	const char* pszFileName, HBITMAP& hbmpLoad,
	bool& bGrayScale)
{

#ifdef FILE_SUPPORT_JPEG

	if (IsFileNameJPEG(pszFileName)) {
		return LoadBitmapFileFromJPEG(
			pszFileName, hbmpLoad,
			bGrayScale);
	}

#endif	// FILE_SUPPORT_JPEG

#ifdef FILE_SUPPORT_PNG

	if (IsFileNamePNG(pszFileName)) {
		return LoadBitmapFileFromPNG(
			pszFileName, hbmpLoad,
			bGrayScale);
	}

#endif	// FILE_SUPPORT_PNG

	return LoadBitmapFileFromBMP(
		pszFileName, hbmpLoad,
		bGrayScale);
}

bool SaveBitmapFile(
	const char* pszFileName, HBITMAP hbmpSave,
	bool bGrayScale, int nQuality)
{

#ifdef FILE_SUPPORT_JPEG

	if (IsFileNameJPEG(pszFileName)) {
		return SaveBitmapFileToJPEG(
			pszFileName, hbmpSave,
			bGrayScale, nQuality);
	}

#endif	// FILE_SUPPORT_JPEG

#ifdef FILE_SUPPORT_PNG

	if (IsFileNamePNG(pszFileName)) {
		return SaveBitmapFileToPNG(
			pszFileName, hbmpSave,
			bGrayScale, nQuality);
	}

#endif	// FILE_SUPPORT_PNG

	return SaveBitmapFileToBMP(
		pszFileName, hbmpSave,
		bGrayScale, nQuality);
}

bool GetClipboardBitmap(
	HWND& hwndTmp, HBITMAP& hbmpGet,
	bool& bGrayScale)
{
	hbmpGet = NULL;
	bGrayScale = false;
	if (!IsClipboardFormatAvailable(CF_DIB)) {
		return false;
	}
	if (hwndTmp == NULL) {
		hwndTmp = CreateTmpWindow();
		if (hwndTmp == NULL) {
			return false;
		}
	}
	if (!OpenClipboard(hwndTmp)) {
		return false;
	}
	HGLOBAL hgDIB = (HGLOBAL)GetClipboardData(CF_DIB);
	if (hgDIB != NULL) {
		BYTE* pbtDIB = (BYTE*)GlobalLock(hgDIB);
		if (pbtDIB != NULL) {
			BITMAPINFO* pbmi = (BITMAPINFO*)pbtDIB;
			int nWidth = pbmi->bmiHeader.biWidth,
				nHeight = pbmi->bmiHeader.biHeight,
				nDepth = pbmi->bmiHeader.biBitCount*
					pbmi->bmiHeader.biPlanes,
				nScanLine = ((nWidth*nDepth+31)/32)*4,
				nBitsLength = nScanLine*nHeight,
				nPalCount = max(
					pbmi->bmiHeader.biClrUsed,
					pbmi->bmiHeader.biClrImportant);
			if ((nPalCount == 0) && (nDepth <= 8)) {
				nPalCount = (1 << nDepth);
			}
			int nInfoLength = sizeof(BITMAPINFOHEADER)+
				((pbmi->bmiHeader.biCompression == BI_BITFIELDS)?
					(sizeof(DWORD)*3): 0)+
				sizeof(RGBQUAD)*nPalCount;
			void* pvBits;
			HDC hdcDesktop = GetDC(HWND_DESKTOP);
			if (hdcDesktop != NULL) {
				hbmpGet = CreateDIBSection(
					hdcDesktop,
					pbmi, DIB_RGB_COLORS,
					&pvBits,
					NULL, 0);
				ReleaseDC(HWND_DESKTOP, hdcDesktop);
			}
			if (hbmpGet != NULL) {
				memcpy(pvBits, pbtDIB+nInfoLength, nBitsLength);
			}
			GlobalUnlock(hgDIB);
		}
	}
	CloseClipboard();
	if (hbmpGet != NULL) {
		bGrayScale = IsBitmapGrayScale(hbmpGet);
	}
	return hbmpGet != NULL;
}

bool SetClipboardBitmap(HWND& hwndTmp, HBITMAP hbmpSet) {
	if (hwndTmp == NULL) {
		hwndTmp = CreateTmpWindow();
		if (hwndTmp == NULL) {
			return false;
		}
	}
	DIBSECTION ds;
	memset(&ds, 0, sizeof(ds));
	if (GetObject(hbmpSet, sizeof(ds), &ds) != sizeof(ds)) {
		return false;
	}
	int nWidth = ds.dsBmih.biWidth,
		nHeight = ds.dsBmih.biHeight,
		nDepth = ds.dsBmih.biBitCount*ds.dsBmih.biPlanes,
		nScanLine = ((nWidth*nDepth+31)/32)*4,
		nBitsLength = nScanLine*nHeight,
		nPalCount = max(
			ds.dsBmih.biClrUsed,
			ds.dsBmih.biClrImportant);
	if ((nPalCount == 0) && (nDepth <= 8)) {
		nPalCount = (1 << nDepth);
	}
	RGBQUAD argbPalettes[256];
	memset(argbPalettes, 0, sizeof(argbPalettes));
	if (nPalCount > 0) {
		HDC hdcMem = CreateCompatibleDC(NULL);
		if (hdcMem != NULL) {
			HBITMAP hbmpOrg = (HBITMAP)SelectObject(hdcMem, hbmpSet);
			GetDIBColorTable(hdcMem, 0, nPalCount, argbPalettes);
			SelectObject(hdcMem, hbmpOrg);
			DeleteDC(hdcMem);
		}
	}
	int nInfoLength = sizeof(BITMAPINFOHEADER)+
		((ds.dsBmih.biCompression == BI_BITFIELDS)?
			(sizeof(DWORD)*3): 0)+
		sizeof(RGBQUAD)*nPalCount;
	HGLOBAL hgDIB = GlobalAlloc(GHND, nInfoLength+nBitsLength);
	if (hgDIB == NULL) {
		return false;
	}
	BYTE* pbtDIB = (BYTE*)GlobalLock(hgDIB);
	if (pbtDIB == NULL) {
		GlobalFree(hgDIB);
		return false;
	}
	memcpy(pbtDIB, &ds.dsBmih, sizeof(BITMAPINFOHEADER));
	pbtDIB += sizeof(BITMAPINFOHEADER);
	if (ds.dsBmih.biCompression == BI_BITFIELDS) {
		memcpy(pbtDIB, ds.dsBitfields, sizeof(DWORD)*3);
		pbtDIB += sizeof(DWORD)*3;
	}
	if (nPalCount > 0) {
		memcpy(pbtDIB, argbPalettes, sizeof(RGBQUAD)*nPalCount);
		pbtDIB += sizeof(RGBQUAD)*nPalCount;
	}
	memcpy(pbtDIB, ds.dsBm.bmBits, nBitsLength);
	GlobalUnlock(hgDIB);
	bool bResult = false;
	if (OpenClipboard(hwndTmp)) {
		if (EmptyClipboard()) {
			SetClipboardData(CF_DIB, hgDIB);
			bResult = true;
		}
		CloseClipboard();
	}
	if (!bResult) {
		GlobalFree(hgDIB);
	}
	return bResult;
}

HBITMAP CreateNewBitmap(
	int nWidth, int nHeight,
	int nDepth, bool bGrayScale)
{
	if ((nWidth < 1) || (nHeight < 1) ||
		((nDepth != 16) && (nDepth != 24) && (nDepth != 32)))
	{
		return NULL;
	}
	if (bGrayScale) {
		nDepth = 8;
	}
	BYTE abtTmp[sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*256];
	memset(abtTmp, 0, sizeof(abtTmp));
	BITMAPINFO* pbmi = (BITMAPINFO*)abtTmp;
	pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	pbmi->bmiHeader.biWidth = nWidth;
	pbmi->bmiHeader.biHeight = nHeight;
	pbmi->bmiHeader.biBitCount = nDepth;
	pbmi->bmiHeader.biPlanes = 1;
	pbmi->bmiHeader.biCompression = BI_RGB;
	if (bGrayScale) {
		for (int i = 0; i < 256; i++) {
			pbmi->bmiColors[i].rgbRed =
				pbmi->bmiColors[i].rgbGreen =
					pbmi->bmiColors[i].rgbBlue = (BYTE)i;
		}
	}
	void* pvBits;
	HDC hdcDesktop = GetDC(HWND_DESKTOP);
	if (hdcDesktop == NULL) {
		return NULL;
	}
	HBITMAP hbmpCreate = CreateDIBSection(
		hdcDesktop,
		pbmi, DIB_RGB_COLORS,
		&pvBits,
		NULL, 0);
	ReleaseDC(HWND_DESKTOP, hdcDesktop);
	return hbmpCreate;
}

bool BitmapExpand(HBITMAP hbmpDst, HBITMAP hbmpSrc) {
	int nSrcWidth, nSrcHeight;
	if (!GetBitmapSize(hbmpSrc, nSrcWidth, nSrcHeight)) {
		return false;
	}
	int nDstWidth, nDstHeight;
	if (!GetBitmapSize(hbmpDst, nDstWidth, nDstHeight)) {
		return false;
	}
	double fZoomX = 0, fZoomY = 0;
	if (nDstWidth > 0) {
		fZoomX = ((double)nSrcWidth)/nDstWidth;
	}
	if (nDstHeight > 0) {
		fZoomY = ((double)nSrcHeight)/nDstHeight;
	}
	HDC hdcDesktop = GetDC(HWND_DESKTOP);
	if (hdcDesktop == NULL) {
		return NULL;
	}
	HDC hdcSrc = CreateCompatibleDC(hdcDesktop),
		hdcDst = CreateCompatibleDC(hdcDesktop);
	ReleaseDC(HWND_DESKTOP, hdcDesktop);
	if ((hdcSrc == NULL) || (hdcDst == NULL)) {
		if (hdcSrc == NULL) {
			DeleteDC(hdcSrc);
		}
		if (hdcDst == NULL) {
			DeleteDC(hdcDst);
		}
	}
	HBITMAP hbmpOrgSrc = (HBITMAP)SelectObject(hdcSrc, hbmpSrc),
		hbmpOrgDst = (HBITMAP)SelectObject(hdcDst, hbmpDst);
	for (int y = 0; y < nDstHeight; y++) {
		for (int x = 0; x < nDstWidth; x++) {
			double fX = (x+0.5)*fZoomX-0.5, fY = (y+0.5)*fZoomY-0.5;
			int x2 = (int)fX, y2 = (int)fY;
			double fOfsX = fmod(fX, 1), fOfsY = fmod(fY, 1);
			if ((x2 < 0) || (fOfsX < 0)) {
				x2 = 0;
				fOfsX = 0;
			} else if (x2 >= nSrcWidth-1) {
				x2 = nSrcWidth-1;
				fOfsX = 0;
			}
			if ((y2 < 0) || (fOfsY < 0)) {
				y2 = 0;
				fOfsY = 0;
			} else if (y2 >= nSrcHeight-1) {
				y2 = nSrcHeight-1;
				fOfsY = 0;
			}
			double fAugX = (cos(fOfsX*PI)+1)/2,
				fAugY = (cos(fOfsY*PI)+1)/2;
			CColor col = GetPixel(hdcSrc, x2, y2);
			if (fOfsX > 0) {
				col = col*fAugX+
					CColor(GetPixel(hdcSrc, x2+1, y2))*(1-fAugX);
			}
			if (fOfsY > 0) {
				CColor col2 = GetPixel(hdcSrc, x2, y2+1);
				if (fOfsX > 0) {
					col2 = col2*fAugX+
						CColor(GetPixel(hdcSrc, x2+1, y2+1))*(1-fAugX);
				}
				col = col*fAugY+col2*(1-fAugY);
			}
			SetPixel(hdcDst, x, y, (COLORREF)col);
		}
	}
	SelectObject(hdcSrc, hbmpOrgSrc);
	SelectObject(hdcDst, hbmpOrgDst);
	DeleteDC(hdcSrc);
	DeleteDC(hdcDst);
	return true;
}

bool BitmapReduce(HBITMAP hbmpDst, HBITMAP hbmpSrc) {
	int nSrcWidth, nSrcHeight;
	if (!GetBitmapSize(hbmpSrc, nSrcWidth, nSrcHeight)) {
		return false;
	}
	int nDstWidth, nDstHeight;
	if (!GetBitmapSize(hbmpDst, nDstWidth, nDstHeight)) {
		return false;
	}
	double fZoomX = 0, fZoomY = 0;
	if (nDstWidth > 0) {
		fZoomX = ((double)nSrcWidth)/nDstWidth;
	}
	if (nDstHeight > 0) {
		fZoomY = ((double)nSrcHeight)/nDstHeight;
	}
	HDC hdcDesktop = GetDC(HWND_DESKTOP);
	if (hdcDesktop == NULL) {
		return NULL;
	}
	HDC hdcSrc = CreateCompatibleDC(hdcDesktop),
		hdcDst = CreateCompatibleDC(hdcDesktop);
	ReleaseDC(HWND_DESKTOP, hdcDesktop);
	if ((hdcSrc == NULL) || (hdcDst == NULL)) {
		if (hdcSrc == NULL) {
			DeleteDC(hdcSrc);
		}
		if (hdcDst == NULL) {
			DeleteDC(hdcDst);
		}
	}
	HBITMAP hbmpOrgSrc = (HBITMAP)SelectObject(hdcSrc, hbmpSrc),
		hbmpOrgDst = (HBITMAP)SelectObject(hdcDst, hbmpDst);
	for (int y = 0; y < nDstHeight; y++) {
		for (int x = 0; x < nDstWidth; x++) {
			double fX1 = x*fZoomX, fY1 = y*fZoomY,
				fX2 = (x+1)*fZoomX, fY2 = (y+1)*fZoomY,
				fX3, fY3;
			CColor col;
			for (double fY = fY1; fY < fY2; fY = fY3) {
				fY3 = floor(fY)+1;
				if (fY3 > fY2) {
					fY3 = fY2;
				}
				double fDY = fY3-fY;
				if (fZoomY > 0) {
					fDY /= fZoomY;
				}
				for (double fX = fX1; fX < fX2; fX = fX3) {
					fX3 = floor(fX)+1;
					if (fX3 > fX2) {
						fX3 = fX2;
					}
					double fDX = fX3-fX;
					if (fZoomX > 0) {
						fDX /= fZoomX;
					}
					int x2 = (int)fX, y2 = (int)fY;
					if (x2 < 0) {
						x2 = 0;
					} else if (x2 > nSrcWidth-1) {
						x2 = nSrcWidth-1;
					}
					col += CColor(GetPixel(hdcSrc, x2, y2))*fDX*fDY;
				}
			}
			SetPixel(hdcDst, x, y, (COLORREF)col);
		}
	}
	SelectObject(hdcSrc, hbmpOrgSrc);
	SelectObject(hdcDst, hbmpOrgDst);
	DeleteDC(hdcSrc);
	DeleteDC(hdcDst);
	return true;
}

int main(int argc, char* argv[]) {
	double fZoomRate = -1;
	int nDstWidth = -1, nDstHeight = -1,
		nDstDepth = 24,
		nDstQuality = 75;
	bool bHelp = false;
	const char* pszSrcFileName = NULL;
	const char* pszDstFileName = NULL;
	const char* pszClipboard = "Clipboard";
	HWND hwndTmp = NULL;
	try {
		for (int i = 1; i < argc; i++) {
			if (argv[i][0] == '-') {
				switch (tolower(argv[i][1])) {
				case 'h':
				case '?':
					bHelp = true;
					break;
				case 'z':
					fZoomRate = atof(argv[i]+2);
					if ((fZoomRate < 1.0) || (fZoomRate > 10000.0)) {
						throw "bad zoom rate";
					}
					break;
				case 'x':
					nDstWidth = atoi(argv[i]+2);
					if ((nDstWidth < 1) || (nDstWidth > 10000)) {
						throw "bad dst width";
					}
					break;
				case 'y':
					nDstHeight = atoi(argv[i]+2);
					if ((nDstHeight < 1) || (nDstHeight > 10000)) {
						throw "bad dst height";
					}
					break;
				case 'd':
					nDstDepth = atoi(argv[i]+2);
					if ((nDstDepth != 16) &&
						(nDstDepth != 24) &&
						(nDstDepth != 32))
					{
						throw "bad dst depth";
					}
					break;
				case 'q':
					nDstQuality = atoi(argv[i]+2);
					if ((nDstQuality < 0) || (nDstQuality > 100)) {
						throw "bad dst quality";
					}
					break;
				case 'c':
					switch (tolower(argv[i][2])) {
					case 'i':
						pszSrcFileName = pszClipboard;
						break;
					case 'o':
						pszDstFileName = pszClipboard;
						break;
					}
					break;
				}
			} else {
				if ((pszSrcFileName == NULL) ||
					(pszDstFileName == NULL))
				{
					char* pszFileName = strdup(argv[i]);
					int nLength = strlen(pszFileName);
					if (((pszFileName[0] == '"') &&
							(pszFileName[nLength-1] == '"')) ||
						((pszFileName[0] == '\'') &&
							(pszFileName[nLength-1] == '\'')))
					{
						if (nLength > 2) {
							memmove(pszFileName, pszFileName+1, nLength-2);
						}
						pszFileName[nLength-2] = '\0';
					}
					if (pszSrcFileName == NULL) {
						pszSrcFileName = pszFileName;
					} else if (pszDstFileName == NULL) {
						pszDstFileName = pszFileName;
					}
				}
			}
		}
		if ((pszSrcFileName == NULL) || (pszDstFileName == NULL)) {
			bHelp = true;
		}
		if (bHelp) {
			xprintf(
				"Usage:\n"
				"\tBmpZoomer [option] <srcfile> <dstfile>\n"
				"\n"
				"SupportFile:\n"
				"\twindows bitmap / jpeg / png\n"
				"\n"
				"Option:\n"
				"\t-h/-?\tshow help(this message)\n"
				"\t-zN\tzoom rate(N%%)\n"
				"\t-xN\tdst width(N dot)\n"
				"\t-yN\tdst height(N dot)\n"
				"\t-dN\tdst bitmap depth(N = 16/24/32, default = 24)\n"
				"\t-qN\tdst jpeg quality(N = 0-100, default = 75)\n"
				"\t-ci\tinput from clipboard\n"
				"\t-co\toutput to clipboard\n"
				"\n"
				"Example:\n"
				"\tBmpZoomer -z50 src.bmp dst.bmp\n"
				"\tBmpZoomer -x64 -y40 src.jpg dst.jpg\n"
				"\tBmpZoomer -z25 -ci dst.png\n"
				"\tBmpZoomer -z150 -ci -co\n");
			return 0;
		}

#ifdef FILE_SUPPORT_JPEG

		if ((pszDstFileName != pszClipboard) &&
			IsFileNameJPEG(pszDstFileName))
		{
			nDstDepth = 24;
		}

#endif	// FILE_SUPPORT_JPEG

#ifdef FILE_SUPPORT_PNG

		if ((pszDstFileName != pszClipboard) &&
			IsFileNamePNG(pszDstFileName))
		{
			nDstDepth = 24;
		}

#endif	// FILE_SUPPORT_PNG

		bool bGrayScale;
		HBITMAP hbmpSrc;
		if (pszSrcFileName != pszClipboard) {
			if (!LoadBitmapFile(
					pszSrcFileName, hbmpSrc,
					bGrayScale))
			{
				throw "cannot load srcfile";
			}
		} else {
			if (!GetClipboardBitmap(
					hwndTmp, hbmpSrc,
					bGrayScale))
			{
				throw "cannot get clipboard";
			}
		}
		int nSrcWidth, nSrcHeight;
		if (!GetBitmapSize(hbmpSrc, nSrcWidth, nSrcHeight)) {
			throw "internal error";
		}
		if (fZoomRate > 0) {
			nDstWidth = (int)(nSrcWidth*fZoomRate/100.0+0.5);
			if (nDstWidth < 1) {
				nDstWidth = 1;
			}
			nDstHeight = (int)(nSrcHeight*fZoomRate/100.0+0.5);
			if (nDstHeight < 1) {
				nDstHeight = 1;
			}
		}
		if (nDstWidth < 0) {
			nDstWidth = nSrcWidth;
		}
		if (nDstHeight < 0) {
			nDstHeight = nSrcHeight;
		}
		xprintf(
			"%s(%dx%d) -> %s(%dx%d)\n",
			pszSrcFileName, nSrcWidth, nSrcHeight,
			pszDstFileName, nDstWidth, nDstHeight);
		if (((nDstWidth > nSrcWidth) && (nDstHeight < nSrcHeight)) ||
			((nDstWidth < nSrcWidth) && (nDstHeight > nSrcHeight)))
		{
			int nTmpWidth = min(nSrcWidth, nDstWidth),
				nTmpHeight = min(nSrcHeight, nDstHeight);
			HBITMAP hbmpTmp = CreateNewBitmap(
				nTmpWidth, nTmpHeight,
				nDstDepth, bGrayScale);
			if (hbmpTmp == NULL) {
				throw "cannot create bitmap";
			}
			if (!BitmapReduce(hbmpTmp, hbmpSrc)) {
				throw "cannot reduce";
			}
			DeleteObject(hbmpSrc);
			hbmpSrc = hbmpTmp;
			nSrcWidth = nTmpWidth;
			nSrcHeight = nTmpHeight;
		}
		HBITMAP hbmpDst = CreateNewBitmap(
			nDstWidth, nDstHeight,
			nDstDepth, bGrayScale);
		if (hbmpDst == NULL) {
			throw "cannot create bitmap";
		}
		if ((nDstWidth >= nSrcWidth) || (nDstHeight >= nSrcHeight)) {
			if (!BitmapExpand(hbmpDst, hbmpSrc)) {
				throw "cannot expand";
			}
		} else {
			if (!BitmapReduce(hbmpDst, hbmpSrc)) {
				throw "cannot reduce";
			}
		}
		if (pszDstFileName != pszClipboard) {
			if (!SaveBitmapFile(
					pszDstFileName, hbmpDst,
					bGrayScale, nDstQuality))
			{
				throw "cannot save dstfile";
			}
		} else {
			if (!SetClipboardBitmap(hwndTmp, hbmpDst)) {
				throw "cannot set clipboard";
			}
		}
		xprintf("Complete.\n");
	} catch (const char* pszError) {
		xprintf("Error : %s\n", pszError);
		return 1;
	}
	return 0;
}

#ifndef _CONSOLE

int WINAPI WinMain(
	HINSTANCE /*hInstance*/, HINSTANCE /*hPrevInstance*/,
	LPSTR /*lpCmdLine*/, int /*nShowCmd*/)
{
	return main(__argc, __argv);
}

#endif
